import aabpl

# def test_aabpl():
#     """
#     test invert_quality_of_life with testdata and default parameters
#     """
#     try:
#         aabpl.invert_quality_of_life(df=aabpl.testdata)
#         success=True
#     except:
#         success=False
#     assert success
    
